use std::cell::RefCell;
use std::fs;
use std::io::ErrorKind;
use std::path::Path;
use std::rc::Rc;

use crate::card::Card;
use crate::db::*;
use crate::ui::cards::new::*;
use crate::widgets::card_entry::CardEntry;
use glib::subclass::InitializingObject;
use gtk::glib::object::ObjectExt;
use gtk::glib::{clone, closure_local};
use gtk::prelude::WidgetExt;
use gtk::subclass::prelude::*;
use gtk::ListBox;
use gtk::{gio, glib, Button, CompositeTemplate, ScrolledWindow, SearchEntry, Window};
use gtk::{prelude::*, FileDialog};
use rusqlite::Connection;
use sha256::try_digest;

#[derive(CompositeTemplate, Default)]
#[template(resource = "/org/foxarmy/learn-hieroglyph/cards/edit/ui.xml")]
pub struct MemoryCardsEditScene {
    #[template_child]
    pub search_entry: TemplateChild<SearchEntry>,
    #[template_child]
    pub add_button: TemplateChild<Button>,
    #[template_child]
    pub cards_container: TemplateChild<ListBox>,
    #[template_child]
    pub cards_scrolled_window: TemplateChild<ScrolledWindow>,
    #[template_child]
    pub back_button: TemplateChild<Button>,

    displaying_cards: RefCell<Vec<CardEntry>>,
}

#[glib::object_subclass]
impl ObjectSubclass for MemoryCardsEditScene {
    const NAME: &'static str = "MemoryCardsEditScene";
    type Type = super::MemoryCardsEditScene;
    type ParentType = gtk::ApplicationWindow;

    fn class_init(klass: &mut Self::Class) {
        klass.bind_template();
    }

    fn instance_init(obj: &InitializingObject<Self>) {
        obj.init_template();
    }
}

impl ObjectImpl for MemoryCardsEditScene {
    fn constructed(&self) {
        self.parent_constructed();

        self.query_cards(None);
        self.update_card_list();

        let binding = self.obj();

        self.add_button.connect_closure("clicked", 
                                        false,
                                        closure_local!(@strong binding => move |_b: &Button| {
                                        let new_win = Rc::new(MemoryCardsNewScene::new(&binding.application().unwrap()));
                                        new_win.get_file_choose_button().connect_clicked(clone!(@strong new_win => move |b: &Button| {
                                            b.set_visible(false);
                                            gtk::glib::MainContext::default().spawn_local(new_card_setup(Rc::clone(&new_win)));
                                        }));
                                        new_win.get_done_button().connect_closure("clicked", true, closure_local!(@strong binding => move |_w: &Button| {
                                            binding.imp().query_cards(None);
                                            binding.imp().update_card_list();
                                        }));
                                        new_win.present();
        }));

        self.search_entry.connect_closure("changed", false, closure_local!(@strong binding => move |e: &SearchEntry| {
            binding.imp().query_cards(
                match e.text().as_str() {
                    "" => None,
                    _ => {
                        let search_option = e.text().to_string();
                        let conditions = format!("hieroglyph LIKE '%{}%' OR reading LIKE '%{}%' OR translation LIKE '%{}%'", search_option, search_option, search_option);
                        Some(conditions)
                    }
                }
            );
            binding.imp().update_card_list();
        }));
        // self.update_card_list();
    }
}

impl MemoryCardsEditScene {
    pub fn clear_displaying_card(&self) {
        let c: &ListBox = self.cards_container.as_ref();

        while c.first_child() != None {
            c.remove(&c.first_child().unwrap());
        }
    }

    pub fn query_cards(&self, options: Option<String>) {
        let conn = Connection::open(get_db_path()).unwrap();

        let selector = match options {
            Some(s) => "WHERE ".to_owned() + &s,
            None => "".to_owned(),
        };

        let sql = format!(
            "SELECT imagename, hieroglyph, reading, translation, is_learning FROM cards {selector}"
        );
        let mut stmt = conn.prepare(sql.as_str()).unwrap();
        let cards_iter = stmt
            .query_map([], |row| {
                let image_path: String = match row.get(0) {
                    Ok(path) => path,
                    Err(_) => String::from(""),
                };
                let c = Card::new(
                    Some(get_images_store_path() + &image_path),
                    Some(row.get(1).unwrap()),
                    Some(row.get(2).unwrap()),
                    Some(row.get(3).unwrap()),
                    Some(row.get(4).unwrap()),
                );
                let entry = CardEntry::new(&c);
                entry.update_state();
                Ok(entry)
            })
            .unwrap();
        self.displaying_cards.borrow_mut().clear();
        for card in cards_iter {
            let card_binding = card.unwrap();
            self.displaying_cards.borrow_mut().push(card_binding);
        }
    }

    pub fn update_card_list(&self) {
        let c: &ListBox = self.cards_container.as_ref();

        self.clear_displaying_card();

        for card in self.displaying_cards.borrow().iter() {
            let self_binding = self.obj();
            c.append(card);

            card.get_delete_button_widget().connect_closure("clicked", false, closure_local!(@strong card, @strong self_binding => move |_b: &Button| {
                let connection = Connection::open(get_db_path()).unwrap();
                let imagepath = &card.imagepath();
                let imagename = &Path::new(&imagepath).file_name().unwrap().to_str().unwrap();
                fs::remove_file(get_images_store_path() + "/" + &imagename).expect("Could not remove file!");
                connection.execute("DELETE FROM cards WHERE imagename = ?1", [&imagename]).unwrap();
                self_binding.imp().query_cards(None);
                self_binding.imp().update_card_list();
            }));

            card.get_edit_button_widget().connect_closure("clicked", false, closure_local!(@strong self_binding as binding, @strong card => move |_b: &Button| {
                let new_win = Rc::new(MemoryCardsNewScene::new(&binding.application().unwrap()));

                //setting corresponding properties
                new_win.get_picture_widget().set_file(Some(&gio::File::for_path(card.imagepath())));
                new_win.get_hieroglyph_entry().set_text(&card.hieroglyph());
                new_win.get_reading_entry().set_text(&card.reading());
                new_win.get_translation_entry().set_text(&card.translation());

                new_win.get_file_choose_button().connect_clicked(clone!(@strong new_win, @strong card => move |b: &Button| {
                    b.set_visible(false);
                    gtk::glib::MainContext::default().spawn_local(new_card_setup(Rc::clone(&new_win)));
                }));
                new_win.get_done_button().connect_closure("clicked", true, closure_local!(@strong binding => move |_w: &Button| {
                    binding.imp().query_cards(None);
                    binding.imp().update_card_list();
                }));
                new_win.present();
            }));
            card.update_file_for_image();
        }
    }
}

async fn new_card_setup<W: IsA<gtk::Window>>(window: Rc<W>) {
    let w: &MemoryCardsNewScene = Into::<&Window>::into(window.upcast_ref())
        .downcast_ref()
        .unwrap(); // Weird casting from &Window as passed in func to &MemoryCardsNewScene

    let picture_widget = w.get_picture_widget();

    let dialog: FileDialog = gtk::FileDialog::builder().build();
    let answer = dialog.open_future(Some(&*window)).await;
    let path = match answer {
        Ok(p) => p,
        Err(_) => return,
    }
    .path()
    .unwrap();
    let path: String = path.as_path().to_str().unwrap().to_owned();

    picture_widget.set_file(Some(&gio::File::for_path(&path)));
    println!("Image path has changed");

    let images_store_path = get_program_home_path() + "/images";

    match fs::create_dir_all(&images_store_path) {
        Ok(_) => {}
        Err(error) => match error.kind() {
            ErrorKind::AlreadyExists => {}
            _ => panic!("Could not create directory for storing images!"),
        },
    };
    w.get_done_button().connect_closure(
        "clicked",
        false,
        closure_local!(@strong w, @strong path => move |_b: &Button| {
            let hash = try_digest(&path).unwrap();
            let extenstion = Path::new(path.as_str()).extension().unwrap().to_str().unwrap();
            let new_filename: String = hash.as_str().to_owned() + "." + extenstion;
            let stored_image_path = Path::new(&images_store_path).join(&new_filename);
            fs::copy(&path, stored_image_path).expect("Error copying image to store");

            let hieroglyph = w.get_hieroglyph_input();
            let reading = w.get_reading_input();
            let translation = w.get_translation_input();
            let conn = Connection::open(get_db_path()).unwrap();

            conn.execute("INSERT OR REPLACE INTO cards ( id, 
                                                imagename, 
                                                hieroglyph, 
                                                reading, 
                                                translation, 
                                                is_learning
                                             ) VALUES
                                             (
                                                (SELECT id FROM cards WHERE hieroglyph = ?2), 
                                                ?1, 
                                                ?2, 
                                                ?3, 
                                                ?4, 
                                                (SELECT is_learning FROM cards WHERE hieroglyph = ?2 
                                                )
                                            )", (&new_filename, &hieroglyph, &reading, &translation)).unwrap();
            conn.execute("UPDATE cards SET is_learning = TRUE WHERE hieroglyph = ?1", [&hieroglyph]).unwrap();
            w.close();

        }),
    );
}

impl WidgetImpl for MemoryCardsEditScene {}

impl WindowImpl for MemoryCardsEditScene {}

impl ApplicationWindowImpl for MemoryCardsEditScene {}
